@@ -26,17 +26,38 @@ class Epsilon:
   """Scheduler class for the regularization parameter epsilon."""
 
   def __init__(self,
-               target: float = 1e-2,
-               init: float = 1.0,
-               decay: float = 1.0):
-    self.target = target
-    self._init = init
-    self._decay = decay
+               target: Optional[float] = None,
+               scale: Optional[float] = None,
+               init: Optional[float] = None,
+               decay: Optional[float] = None):
+    """Initializes a scheduler using possibly geometric decay.
+
+    The entropic regularization value is given either directly or relative to
+    a scale. In that case, the initial ``target`` value is understood to be a
+    proportion of the ``scale``. Both are recorded and merged in the ``target``
+    field which is built from those two parameters.
+
+    Args:
+      target: the epsilon regularizing value that is targeted, understood
+        as a multiple of scale.
+      scale: scale to be used with target_init to define a target epsilon.
+      init: initial value when using epsilon scheduling, understood as a
+        a fraction of scale as well.
+      decay: geometric decay factor, smaller than 1.
+    """
+    self._target_init = .01 if target is None else target
+    self._scale = 1.0 if scale is None else scale
+    self._init = 1.0 if init is None else init
+    self._decay = 1.0 if decay is None else decay
+
+  @property
+  def target(self):
+    return self._target_init * self._scale
 
   def at(self, iteration: Optional[int] = 1) -> float:
     if iteration is None:
       return self.target
-    init = jnp.where(self._decay < 1.0, self._init, self.target)
+    init = jnp.where(self._decay < 1.0, self._init * self._scale, self.target)
     decay = jnp.where(self._decay < 1.0, self._decay, 1.0)
     return jnp.maximum(init * decay**iteration, self.target)
 
@@ -47,7 +68,7 @@ class Epsilon:
     return self.done(self.at(iteration))
 
   def tree_flatten(self):
-    return (self.target, self._init, self._decay), None
+    return (self._target_init, self._scale, self._init, self._decay), None
 
   @classmethod
   def tree_unflatten(cls, aux_data, children):
